@@ -6,6 +6,7 @@ use Data::Dumper;
 use DBI;
 use DBI::Const::GetInfoType;
 use Digest::MD5;
+use Encode;
 use Fcntl qw(:DEFAULT :flock);
 use File::Copy 'move';
 use HTML::Entities;
@@ -13,6 +14,7 @@ use HTTP::Date;
 use MIME::Base64;
 use Number::Format;
 use Time::HiRes 'time';
+use Scalar::Util;
 use Storable;
 
 ################################################################################
@@ -34,6 +36,7 @@ sub check_constants {
 	our @_OVERRIDING_PARAMETER_NAMES     = qw (select __no_navigation __tree __last_query_string);
 
 	our %INC_FRESH = ();
+	our %INC_FRESH_BY_PATH = ();
 
 }
 
@@ -128,7 +131,7 @@ sub check_version {
 
 	require Date::Calc;
 
-	return if $ENV {ELUDIA_BANNER_PRINTED};
+	return if $Eludia::VERSION ||= $ENV {ELUDIA_BANNER_PRINTED};
 	
 	my ($year) = Date::Calc::Today ();
 	
@@ -173,7 +176,7 @@ sub check_version {
 
 EOT
 
-	$ENV {ELUDIA_BANNER_PRINTED} = 1;
+	$ENV {ELUDIA_BANNER_PRINTED} = $Eludia::VERSION;
 
 }
 
@@ -684,31 +687,6 @@ sub check_module_presentation_tools {
 
 ################################################################################
 
-sub check_module_mac {
-
-	print STDERR " check_module_mac.................... ";
-
-	exists $preconf -> {core_no_log_mac} or $preconf -> {core_no_log_mac} = 1;
-	
-	if ($preconf -> {core_no_log_mac}) { 
-
-		eval 'sub get_mac {""}';
-		
-		print STDERR "no MAC logging, ok.\n";
-
-	} 
-	else { 
-
-		require Eludia::Content::Mac;
-
-		print STDERR "MAC logging enabled, ok.\n";		
-
-	}
-
-}
-
-################################################################################
-
 sub check_module_session_access_logs {
 
 	print STDERR " check_module_session_access_logs.... ";
@@ -879,6 +857,61 @@ sub check_module_queries {
 
 ################################################################################
 
+sub check_module_log {
+
+	print STDERR " check_module_log.................... ";
+	
+	$conf -> {core_log} -> {version} ||= 'v1';
+	
+	$preconf -> {core_log} -> {suppress} ||= {
+	
+		always => [qw (
+		
+			__infty
+			__last_query_string
+			__last_scrollable_table_row
+			__no_infty
+			__no_navigation
+			__popup
+			__this_query_string
+			redirect_params
+			sid
+			lang
+			salt
+
+		)],
+		
+		empty  => [qw (
+		
+			__form_checkboxes
+			__suggest
+			__tree
+			select
+			
+		)],
+	
+	};
+	
+	$preconf -> {_} -> {core_log} = $conf -> {core_log};
+
+	!exists $preconf -> {core_no_log_mac}
+
+		or $preconf -> {core_no_log_mac}
+
+			or $preconf -> {_} -> {core_log} -> {log_mac} = 1;
+
+	require "Eludia/Content/Log/$preconf->{_}->{core_log}->{version}.pm";
+	
+	print STDERR "$preconf->{_}->{core_log}->{version}, ok.\n";
+
+}
+
+#############################################################################
+
+sub darn ($) {warn Dumper ($_[0]); return $_[0]}
+
+################################################################################
+
 BEGIN {
 
 	foreach (grep {/^Eludia/} keys %INC) { delete $INC {$_} }
