@@ -1,18 +1,92 @@
 no warnings;
 
+use Eludia::SQL::Transfer;
+
+use Eludia::SQL::TheSqlFunction;
+
+################################################################################
+
+sub lc_hashref {}
+
+################################################################################
+
+sub sql_sessions_timeout_in_minutes {
+
+	my $timeout = $preconf -> {session_timeout} || $conf -> {session_timeout} || 30;
+
+	if ($preconf -> {core_auth_cookie} =~ /^\+(\d+)([mhd])/) {
+	
+		$timeout = $1 * (
+
+			$2 eq 'h' ? 60 :
+			
+			$2 eq 'd' ? 1440 :
+				
+			1
+				
+		)
+	
+	}
+	
+	return $timeout;
+
+}
+
+################################################################################
+
+sub add_vocabularies {
+
+	my ($item, @items) = @_;
+
+	while (@items) {
+	
+		my $name = shift @items;
+		
+		my $options = {};
+		
+		if (@items > 0 && ref $items [0] eq HASH) {
+		
+			$options = shift @items;
+		
+		}
+		
+		next
+			if $options -> {off};
+		
+		$options -> {item} = $item;
+		
+		my $table_name = $options -> {name} || $name;
+		
+		$item -> {$name} = sql_select_vocabulary ($table_name, $options);
+		
+		if ($options -> {ids}) {
+			
+			ref $options -> {ids} eq HASH or $options -> {ids} = {table => $options -> {ids}};
+			
+			$options -> {ids} -> {from}  ||= 'id_' . en_unplural ($_REQUEST {type});
+			$options -> {ids} -> {to}    ||= 'id_' . en_unplural ($table_name);
+			
+			$options -> {ids} -> {name}  ||= $options -> {ids} -> {to};
+			
+			$_REQUEST {"__checkboxes_$options->{ids}->{to}"} = $options -> {ids} -> {table};
+		
+			$item -> {$options -> {ids} -> {name}} = [sql_select_col ("SELECT $options->{ids}->{to} FROM $options->{ids}->{table} WHERE fake = 0 AND $options->{ids}->{from} = ?", $item -> {id})];
+		
+		}
+		
+	}
+	
+	return $item;
+
+}
+
 ################################################################################
 
 sub sql_weave_model {
 
 	my ($db_model) = @_;
 
-	my @tables = ();
-	
-	foreach my $table_name ($db -> tables) {	
-		$table_name =~ s{.*?(\w+)\W*$}{$1}gsm;
-		next if $table_name eq $conf -> {systables} -> {log};
-		push @tables, lc $table_name;
-	}
+	my @tables = grep {$_ ne $conf -> {systables} -> {log}} map {lc} get_tables ();
 		
 	foreach my $table_name (@tables) {
 	
@@ -82,9 +156,10 @@ sub sql_weave_model {
 
 sub check_systables {
 
-	foreach (qw(	
-		__voc_replacements	
-		__access_log		
+	foreach (qw(
+		__access_log
+		__queries
+		__defaults
 		__benchmarks		
 		__request_benchmarks
 		__last_update		
@@ -92,6 +167,7 @@ sub check_systables {
 		__required_files	
 		__screenshots		
 		cache_html		
+		holidays
 		log			
 		roles			
 		sessions		
@@ -113,231 +189,30 @@ sub sql_assert_core_tables {
 	return if $model_update -> {core_ok};
 
 my $time = time;
+
+	$model_update -> assert (
 	
-	if ($conf -> {core_voc_replacement_use}) {
-	
-		$model_update -> assert (tables => {$conf -> {systables} -> {__voc_replacements} => {
+		tables => {
 
-			columns => {
-				id          => {TYPE_NAME => 'bigint', _EXTRA => 'auto_increment', _PK => 1},
-				table_name  => {TYPE_NAME => 'varchar', COLUMN_SIZE => 255},
-				object_name => {TYPE_NAME => 'varchar', COLUMN_SIZE => 255},
-				object_type => {TYPE_NAME => 'int', COLUMN_SIZE => 1},
-			},
-			
-			keys => {
-				ix => 'table_name',
-				ix2 => 'object_name',
-			},
+			$conf -> {systables} -> {__last_update} => {
 
-		}});
-	
-	}
-
-$time = __log_profilinig ($time, ' <sql_assert_core_tables>: 136');
-
-	my %defs = (
-	
-		$conf -> {systables} -> {__access_log} => {
-		
-			columns => {
-				id         => {TYPE_NAME => 'bigint', _EXTRA => 'auto_increment', _PK => 1},
-				id_session => {TYPE_NAME => 'bigint'},
-				ts         => {TYPE_NAME => 'timestamp'},
-				no         => {TYPE_NAME => 'int'},
-				href       => {TYPE_NAME => 'text'},
-			},
-			
-			keys => {
-				ix => 'id_session,no',
-				ix2 => 'id_session,href(255)',
-			},
-
-		},
-		
-		$conf -> {systables} -> {__moved_links} => {
-		
-			columns => {
-				id          => {TYPE_NAME => 'bigint', _EXTRA => 'auto_increment', _PK => 1},
-				table_name  => {TYPE_NAME => 'varchar', COLUMN_SIZE => 255},
-				column_name => {TYPE_NAME => 'varchar', COLUMN_SIZE => 255},
-				id_from     => {TYPE_NAME => 'int'},
-				id_to       => {TYPE_NAME => 'int'},
-			},
-			
-			keys => {
-				id_to => 'id_to',
-			},
-
-		},
-		
-		$conf -> {systables} -> {__required_files} => {
-		
-			columns => {
-				unix_ts   => {TYPE_NAME => 'bigint'},
-				file_name => {TYPE_NAME => 'varchar', COLUMN_SIZE  => 255},
-			},
-			
-			keys => {
-				ix => 'file_name',
-			},
-
-		},
-
-		$conf -> {systables} -> {__last_update} => {
-		
-			columns => {
-				pid 	  => {TYPE_NAME => 'int'},
-				unix_ts   => {TYPE_NAME => 'bigint'},
-			},
-			
-		},
-
-		$conf -> {systables} -> {sessions} => {
-		
-			columns => {
-
-				id      => {TYPE_NAME  => 'bigint', _PK    => 1},
-				id_user => {TYPE_NAME  => 'int'},
-				id_role => {TYPE_NAME  => 'int'},
-				ts      => {TYPE_NAME  => 'timestamp'},
-
-				ip =>     {TYPE_NAME => 'varchar', COLUMN_SIZE => 255},
-				ip_fw =>  {TYPE_NAME => 'varchar', COLUMN_SIZE => 255},
-	
-				peer_server => {TYPE_NAME    => 'varchar', COLUMN_SIZE  => 255},
-				peer_id => {TYPE_NAME    => 'bigint'},
+				columns => {
 				
-			}
-
-		},
-
-		$conf -> {systables} -> {roles} => {
-
-			columns => {
-				id   => {TYPE_NAME  => 'int', _EXTRA => 'auto_increment', _PK => 1},
-				fake => {TYPE_NAME  => 'bigint', COLUMN_DEF => 0, NULLABLE => 0},
-				name  => {TYPE_NAME    => 'varchar', COLUMN_SIZE  => 255},
-				label => {TYPE_NAME    => 'varchar', COLUMN_SIZE  => 255},
-			},
-
-		},
-
-		$conf -> {systables} -> {users} => {
-
-			columns => {
-				id   => {TYPE_NAME  => 'int', _EXTRA => 'auto_increment', _PK => 1},
-				fake => {TYPE_NAME  => 'bigint', COLUMN_DEF => 0, NULLABLE => 0},
-				name =>     {TYPE_NAME => 'varchar', COLUMN_SIZE => 255},
-				login =>    {TYPE_NAME => 'varchar', COLUMN_SIZE => 255},
-				label =>    {TYPE_NAME => 'varchar', COLUMN_SIZE => 255},
-				password => {TYPE_NAME => 'varchar', COLUMN_SIZE => 255},
-				mail     => {TYPE_NAME => 'varchar', COLUMN_SIZE => 255},
-				id_role  =>  {TYPE_NAME => 'int'},
-
-				peer_server => {TYPE_NAME    => 'varchar', COLUMN_SIZE  => 255},
-				peer_id => {TYPE_NAME    => 'int'},
+					id        => {TYPE_NAME => 'bigint', _EXTRA => 'auto_increment', _PK => 1},
+					pid 	  => {TYPE_NAME => 'int'},
+					unix_ts   => {TYPE_NAME => 'bigint'},
 				
-				subset => {TYPE_NAME    => 'varchar', COLUMN_SIZE  => 255},
+				},
 
-			}
-
-		},
-
-		$conf -> {systables} -> {log} => {
-
-			columns => {
-				id   => {TYPE_NAME  => 'int', _EXTRA => 'auto_increment', _PK => 1},
-				fake => {TYPE_NAME  => 'bigint', COLUMN_DEF => 0, NULLABLE => 0},
-				id_user =>   {TYPE_NAME => 'int'},
-				id_object => {TYPE_NAME => 'int'},
-				ip =>     {TYPE_NAME => 'varchar', COLUMN_SIZE => 255},
-				ip_fw =>  {TYPE_NAME => 'varchar', COLUMN_SIZE => 255},
-				type =>   {TYPE_NAME => 'varchar', COLUMN_SIZE => 255},
-				action => {TYPE_NAME => 'varchar', COLUMN_SIZE => 255},
-				error  => {TYPE_NAME => 'varchar', COLUMN_SIZE => 255},
-				params => {TYPE_NAME => 'longtext'},
-				dt     => {TYPE_NAME => 'timestamp'},
-				mac    => {TYPE_NAME  => 'varchar', COLUMN_SIZE => 17},
-			}
-
-		},	
-	
+			},
+		
+		}, 
+				
+		prefix => 'sql_assert_core_tables#',
+		
 	);
-	
-	$conf -> {core_cache_html} and $defs {$conf -> {systables} -> {cache_html}} = {
 
-		columns => {
-			uri     => {TYPE_NAME  => 'varchar', COLUMN_SIZE  => 255, _PK    => 1},
-			ts      => {TYPE_NAME  => 'timestamp'},
-		}
-
-	};
-	
-	$preconf -> {core_debug_profiling} > 1 and $defs {$conf->{systables}->{__benchmarks}} = {
-
-		columns => {
-			id       => {TYPE_NAME  => 'int'    , _EXTRA => 'auto_increment', _PK => 1},
-			fake     => {TYPE_NAME  => 'bigint' , COLUMN_DEF => 0, NULLABLE => 0},
-			label    => {TYPE_NAME  => 'varchar', COLUMN_SIZE  => 255},
-			cnt      => {TYPE_NAME  => 'bigint' , COLUMN_DEF => 0, NULLABLE => 0},
-			ms       => {TYPE_NAME  => 'bigint' , COLUMN_DEF => 0, NULLABLE => 0},
-			mean     => {TYPE_NAME  => 'bigint' , COLUMN_DEF => 0, NULLABLE => 0},
-			selected => {TYPE_NAME  => 'bigint' , COLUMN_DEF => 0, NULLABLE => 0},
-			mean_selected => {TYPE_NAME  => 'bigint' , COLUMN_DEF => 0, NULLABLE => 0},
-		},
-		
-		keys => {
-			label => 'label',
-		},
-		
-	};
-
-	$preconf -> {core_debug_profiling} > 2 and $defs {$conf->{systables}->{__request_benchmarks}} = {
-
-		columns => {
-			id	=> {TYPE_NAME  => 'int'    , _EXTRA => 'auto_increment', _PK => 1},
-			fake	=> {TYPE_NAME  => 'bigint' , COLUMN_DEF => 0, NULLABLE => 0},
-			id_user	=> {TYPE_NAME => 'int'},
-			dt	=> {TYPE_NAME => 'timestamp'},
-			params	=> {TYPE_NAME => 'longtext'},
-			ip =>     {TYPE_NAME => 'varchar', COLUMN_SIZE => 255},
-			ip_fw =>  {TYPE_NAME => 'varchar', COLUMN_SIZE => 255},
-			type =>   {TYPE_NAME => 'varchar', COLUMN_SIZE => 255},
-			mac    => {TYPE_NAME  => 'varchar', COLUMN_SIZE => 17},
-
-			connection_id		=> {TYPE_NAME => 'int'},
-			connection_no		=> {TYPE_NAME => 'int'},
-
-			request_time		=> {TYPE_NAME => 'int'},
-			application_time	=> {TYPE_NAME => 'int'},
-			sql_time		=> {TYPE_NAME => 'int'},
-			response_time		=> {TYPE_NAME => 'int'},
-			
-			bytes_sent		=> {TYPE_NAME => 'int'},
-			is_gzipped		=> {TYPE_NAME => 'tinyint'}, 
-		},
-
-	};
-
-	$conf -> {core_screenshot} and $defs {$conf -> {systables} -> {__screenshots}} = {
-
-		columns => {
-			id        => {TYPE_NAME  => 'int', _EXTRA => 'auto_increment', _PK => 1},
-			subset    => {TYPE_NAME => 'varchar', COLUMN_SIZE  => 255},
-			type      => {TYPE_NAME => 'varchar', COLUMN_SIZE  => 255},
-			id_object => {TYPE_NAME => 'int'},
-			id_user	  => {TYPE_NAME => 'int'},
-			dt        => {TYPE_NAME => 'timestamp'},
-			html      => {TYPE_NAME => 'text'},
-			error     => {TYPE_NAME => 'tinyint', COLUMN_DEF => 0},
-			gziped    => {TYPE_NAME => 'tinyint', COLUMN_DEF => 0},
-			params    => {TYPE_NAME => 'text'},
-		},
-
-	};
-
-	$model_update -> assert (tables => \%defs, core_voc_replacement_use => $conf -> {core_voc_replacement_use});
+	sql_version ();
 
 	$model_update -> {core_ok} = 1;
 		
@@ -424,21 +299,41 @@ sub sql_is_temporal_table {
 
 ################################################################################
 
+sub sql_ping {
+
+	my $r;
+
+	eval {
+	
+		my $st = $db -> prepare ('SELECT 1');
+		
+		$st -> execute;
+		
+		$r = $st -> fetchrow_arrayref;
+	
+	};
+	
+	return @$r == 1 && $r -> [0] == 1 ? 1 : 0;
+
+}
+
+################################################################################
+
 sub sql_reconnect {
 
 my $time = time;
 
-	if ($db && $model_update && $model_update -> {core_ok}) {
-		my $ping = $db -> ping;
-$time = __log_profilinig ($time, '  sql_reconnect: ping');
-		return if $ping;
+	our $db, $model_update, $SQL_VERSION;
+
+	if ($db && ($preconf -> {no_model_update} || ($model_update && $model_update -> {core_ok}))) {
+
+		sql_ping () and return
+		
+$time = __log_profilinig ($time, '  sql_reconnect: ping OK');
+
 	}
 	
-	check_systables ();
-
-$time = __log_profilinig ($time, '  sql_reconnect: check_systables');
-
-	our $db = DBI -> connect ($preconf -> {'db_dsn'}, $preconf -> {'db_user'}, $preconf -> {'db_password'}, {
+	$db = DBI -> connect ($preconf -> {db_dsn}, $preconf -> {db_user}, $preconf -> {db_password}, {
 		RaiseError  => 1, 
 		AutoCommit  => 1,
 		LongReadLen => 1000000,
@@ -446,48 +341,77 @@ $time = __log_profilinig ($time, '  sql_reconnect: check_systables');
 		InactiveDestroy => 0,
 	});
 
-$time = __log_profilinig ($time, '  sql_reconnect: connect');
+	if ($preconf -> {db_cache_statements}) {
 
-	my $driver_name = $db -> get_info ($GetInfoType {SQL_DBMS_NAME});
-	
-	$driver_name =~ s{\W}{}gsm;
-
-	eval "require Eludia::SQL::$driver_name";
-
-	die $@ if $@;
-	
-	our $SQL_VERSION = sql_version ();
-	$SQL_VERSION -> {driver} = $driver_name;
-
-	delete $INC {"Eludia/SQL/${driver_name}.pm"};
-
-$time = __log_profilinig ($time, '  sql_reconnect: driver reloaded');
-
-	unless ($preconf -> {no_model_update}) {
-	
-		our $model_update = DBIx::ModelUpdate -> new (		
-			$db, 
-			dump_to_stderr 		=> 1,
-			before_assert		=> $conf -> {'db_temporality'} ? \&sql_temporality_callback : undef,
-			schema			=> $preconf -> {db_schema},
-			_db_model_checksums	=> $conf -> {systables} -> {_db_model_checksums}, 
-			__voc_replacements	=> $conf -> {systables} -> {__voc_replacements}, 
-			core_voc_replacement_use=> $conf -> {core_voc_replacement_use},
-		);
-
-#		$preconf -> {no_model_update} = 1;
+		require Eludia::Content::Tie::LRUHash;
 		
+		my %cache;
+
+		tie %cache, 'Eludia::Content::Tie::LRUHash', {size => 300};
+
+		$db -> {CachedKids} = \%cache;
+
 	}
 
+$time = __log_profilinig ($time, "  sql_reconnect: connected to $preconf->{db_dsn}");
+
+	unless ($INC_FRESH {db_driver}) {
+
+		my $driver_name = $db -> get_info ($GetInfoType {SQL_DBMS_NAME});
+	
+		$driver_name =~ s{\W}{}gsm;
+
+		my $path = __FILE__;
+	
+		$path =~ s{(.)SQL\.pm$}{${1}SQL$1Dialect$1${driver_name}.pm};
+
+		do $path;
+		
+		die $@ if $@;
+
+		$INC_FRESH {db_driver} = time;
+
+		$SQL_VERSION = {driver => $driver_name};
+
+$time = __log_profilinig ($time, "  sql_reconnect: $driver_name is loaded");
+
+	}
+	
+	sql_version ();
+
+$time = __log_profilinig ($time, "  sql_reconnect: driver version is $SQL_VERSION->{string}");
+
+	unless ($preconf -> {no_model_update}) {
+		
+		if ($model_update) {
+		
+			$model_update -> {db} = $db;
+		
+		}
+		else {
+	
+			$model_update = $_NEW_PACKAGE -> new (
+				$db, 
+				before_assert		=> $conf -> {'db_temporality'} ? \&sql_temporality_callback : undef,
+				schema			=> $preconf -> {db_schema},
+			);
+
 $time = __log_profilinig ($time, '  sql_reconnect: $model_update created');
 
+		}
+	
+	}
+
 }   	
 
 ################################################################################
 
 sub sql_disconnect {
-	if ($db) { $db -> disconnect; }
+
+	eval { $db -> disconnect };
+
 	undef $db;
+
 }
 
 ################################################################################
@@ -517,6 +441,100 @@ sub sql_select_vocabulary {
 	}
 	
 	$filter .= " AND $options->{filter}" if $options -> {filter};
+	
+	my @params = ();
+	
+	if ($options -> {in}) {
+	
+		my $in = $options -> {in};
+		
+		my $ref = ref $in;
+	
+		if ($ref eq SCALAR) {
+		
+			my $tied = tied $$in;
+		
+			if (_sql_ok_subselects ()) {
+				
+				$filter .= " AND id IN ($tied->{sql})";
+
+				push @params, @{$tied -> {params}};
+				
+			}
+			else {
+
+				$filter .= " AND id IN ($$in)";
+
+			}
+
+		}
+		elsif ($ref eq ARRAY) {
+		
+			@$in > 0 or return [];
+			
+			$in = join ',', @$in;
+			
+			$filter .= " AND id IN ($in)";
+		
+		}
+		elsif (!$ref) {
+		
+			$in =~ /\d/ or return [];
+
+			$filter .= " AND id IN ($in)";
+		
+		}
+		else {
+			die "Wrong IN list";
+		}
+	
+	}
+
+	if ($options -> {not_in}) {
+	
+		my $in = $options -> {not_in};
+		
+		my $ref = ref $in;
+	
+		if ($ref eq SCALAR) {
+		
+			my $tied = tied $$in;
+		
+			if (_sql_ok_subselects ()) {
+				
+				$filter .= " AND id NOT IN ($tied->{sql})";
+
+				push @params, @{$tied -> {params}};
+				
+			}
+			else {
+
+				$filter .= " AND id NOT IN ($$in)";
+
+			}
+
+		}
+		elsif ($ref eq ARRAY) {
+		
+			@$in > 0 or return [];
+			
+			$in = join ',', @$in;
+			
+			$filter .= " AND id NOT IN ($in)";
+		
+		}
+		elsif (!$ref) {
+		
+			$in =~ /\d/ or return [];
+
+			$filter .= " AND id NOT IN ($in)";
+		
+		}
+		else {
+			die "Wrong [NOT] IN list";
+		}
+	
+	}
 
 	if ($preconf -> {subset} && $table_name eq $conf -> {systables} -> {roles}) {
 		
@@ -541,23 +559,25 @@ sub sql_select_vocabulary {
 	}
 	
 	$options -> {label} .= ', parent' if $options -> {tree};
+	
+	my @list;
+	
+	tie @list, 'Eludia::Tie::Vocabulary', {
+	
+		sql      => "SELECT id, $$options{label}, fake FROM $table_name WHERE $filter ORDER BY $$options{order} $limit",
 		
-	my $list = sql_select_all ("SELECT id, $$options{label} FROM $table_name WHERE $filter ORDER BY $$options{order} $limit");
-	
-	if ($options -> {tree}) {
-	
-		$list = tree_sort ($list);
+		params   => \@params,
 		
-		if (!$_REQUEST {__read_only} || $_REQUEST {__only_form}) {
-
-			foreach (@$list) { $_ -> {label} = ('&nbsp;&nbsp;' x $_ -> {level}) . $_ -> {label} }
-	
-		}
+		_REQUEST => \%_REQUEST,
 		
-	}
-
-	return $list;
-	
+		package  => current_package (),
+		
+		tree     => $options -> {tree},
+		
+	};
+		
+	return \@list;
+			
 }
 
 ################################################################################
@@ -565,6 +585,10 @@ sub sql_select_vocabulary {
 sub sql_select_id {
 
 	my ($table, $values, @lookup_field_sets) = @_;
+	
+	my $result = {};
+
+	my $table_safe = sql_table_name ($table);
 
 	my %values = ();
 	
@@ -584,30 +608,66 @@ sub sql_select_id {
 	
 	@lookup_field_sets = (['label']) if @lookup_field_sets == 0;
 	
+	my $options = ref $lookup_field_sets [-1] eq HASH ? pop @lookup_field_sets : {};
+	
 	my $record = {};
 	
-	foreach my $lookup_fields (@lookup_field_sets) {
+	my $auto_commit = $db -> {AutoCommit};
+	
+	eval { $db -> {AutoCommit} = 0; };
+	
+	sql_lock ($table);
 
-		my $sql = "SELECT * FROM $table WHERE fake <= 0";
+	eval {
+
+	foreach my $lookup_fields (@lookup_field_sets) {
+	
+		if (ref $lookup_fields eq CODE) {		
+			next if &$lookup_fields ();
+			return 0;		
+		}
+
+		my $sql = "SELECT * FROM $table_safe WHERE fake <= 0";
 		my @params = ();
 
 		foreach my $lookup_field (@$lookup_fields) {
-			$sql .= " AND $lookup_field = ?";
-			push @params, $values -> {$lookup_field};
+		
+			my $value = $values -> {$lookup_field};
+			
+			if ($value eq '' && $SQL_VERSION -> {driver} eq 'Oracle') {
+			
+				$value = undef;
+			
+			}
+			
+			if (defined $value) {
+			
+				$sql .= " AND $lookup_field = ?";
+				push @params, $values -> {$lookup_field};
+				
+			}
+			else {
+
+				$sql .= " AND $lookup_field IS NULL";
+
+			}
+		
 		}
 
 		$sql .= " ORDER BY fake DESC, id DESC";
 		
 		$record = sql_select_hash ($sql, @params);
-		
+
 		last if $record -> {id};
 
 	}
 		
-	while (my $id = ($record -> {is_merged_to} || $record -> {id_merged_to})) {
-		$record = sql_select_hash ($table, $id);
+	unless ($_REQUEST {_no_search_merged_record}) {
+		while (my $id = ($record -> {is_merged_to} || $record -> {id_merged_to})) {
+			$record = sql_select_hash ($table, $id);
+		}
 	}
-	
+
 	if ($record -> {id}) {
 	
 		my @keys   = ();
@@ -617,6 +677,8 @@ sub sql_select_id {
 
 			($forced -> {$key} && $values -> {$key} ne $record -> {$key}) or $record -> {$key} eq '' or next;
 
+			$result -> {update} -> {$key} = {old => $record -> {$key}, new => $values -> {$key}};
+
 			push @keys,   $key;
 			push @values, $values -> {$key};
 
@@ -624,13 +686,35 @@ sub sql_select_id {
 
 		if (@keys) {
 
-			sql_do ('UPDATE ' . $table . ' SET ' . (join ', ', map {"$_ = ?"} @keys) . ' WHERE id = ?', @values, $record -> {id});
+			sql_do ('UPDATE ' . $table_safe . ' SET ' . (join ', ', map {"$_ = ?"} @keys) . ' WHERE id = ?', @values, $record -> {id});
 
 		}
 	
 	}
+	
+	unless ($record -> {id}) {
+	
+		$record -> {id} = sql_do_insert ($table, $values);
+		
+		$result -> {insert} = $values;
+	
+	}
 
-	return $record -> {id} || sql_do_insert ($table, $values);
+	
+	};
+
+	sql_unlock ($table);
+	
+	if ($auto_commit) {
+	
+		eval { 
+			$db -> commit;
+			$db -> {AutoCommit} = 1; 
+		};
+
+	}
+	
+	return $options -> {show_diff} && wantarray ? ($record -> {id}, $result) : $record -> {id};
 
 }
 
@@ -638,12 +722,10 @@ sub sql_select_id {
 
 sub sql_do_relink {
 
-	my ($table_name, $old_ids, $new_id) = @_;
+	my ($table_name, $old_ids, $new_id, $options) = @_;			
 	
 	sql_weave_model ($DB_MODEL);
 
-#warn Dumper ($DB_MODEL -> {aliases} -> {$table_name});
-	
 	ref $old_ids eq ARRAY or $old_ids = [$old_ids];
 	
 	my $column_name = '';
@@ -653,6 +735,7 @@ sub sql_do_relink {
 	my $record = sql_select_hash ($table_name, $new_id);
 	my @empty_fields = ();
 	foreach my $key (keys %$record) {
+		next if $options -> {no_update};
 		next if $record -> {$key} . '' ne '';
 		next if $key eq 'id';
 		next if $key eq 'fake';
@@ -660,10 +743,9 @@ sub sql_do_relink {
 		next if $key eq 'id_merged_to';
 		push @empty_fields, $key;
 	}
-		
-#warn Dumper ($DB_MODEL -> {tables} -> {$table_name});
-#warn Dumper ($DB_MODEL -> {tables} -> {$table_name} -> {references});
-
+			
+	my $moved_links_table = sql_table_name ($conf -> {systables} -> {__moved_links});
+				
 	foreach my $old_id (@$old_ids) {
 	
 warn "relink $table_name: $old_id -> $new_id";
@@ -676,12 +758,15 @@ warn "relink $table_name: $old_id -> $new_id";
 
 		foreach my $column_def (@{$DB_MODEL -> {aliases} -> {$table_name} -> {references}}) {
 
+			next
+				if $DB_MODEL -> {tables} -> {$column_def -> {table_name}} -> {sql};
+			
 warn "relink $$column_def{table_name} ($$column_def{name}): $old_id -> $new_id";
 
 			if ($column_def -> {TYPE_NAME} =~ /int/) {
 			
 				sql_do (<<EOS, $old_id);
-					INSERT INTO $conf->{systables}->{__moved_links}
+					INSERT INTO $moved_links_table
 						(table_name, column_name, id_from, id_to)
 					SELECT
 						'$$column_def{table_name}' AS table_name,
@@ -703,7 +788,7 @@ EOS
 				my $_new_id = ',' . $new_id . ',';
 			
 				sql_do (<<EOS, '%' . $old_id . '%');
-					INSERT INTO $conf->{systables}->{__moved_links}}
+					INSERT INTO $moved_links_table
 						(table_name, column_name, id_from, id_to)
 					SELECT
 						'$$column_def{table_name}' AS table_name,
@@ -744,9 +829,11 @@ sub sql_undo_relink {
 	sql_weave_model ($DB_MODEL);
 
 	my ($table_name, $old_ids) = @_;
-	
+
 	ref $old_ids eq ARRAY or $old_ids = [$old_ids];
-			
+
+	my $moved_links_table = sql_table_name ($conf -> {systables} -> {__moved_links});
+	
 	foreach my $old_id (@$old_ids) {
 		
 		$old_id > 0 or next;
@@ -759,7 +846,7 @@ warn "undo relink $table_name: $old_id";
 
 			my $from = <<EOS;
 				FROM
-					$conf->{systables}->{__moved_links}
+					$moved_links_table
 				WHERE
 					table_name = '$$column_def{table_name}'
 					AND column_name = '$$column_def{name}'
@@ -790,19 +877,43 @@ warn "undo relink $$column_def{table_name} ($$column_def{name}): $old_id";
 
 sub assert_fake_key {
 
-	$DB_MODEL -> {tables} -> {$table_name} or return;
-	
 	my ($table_name) = @_;
 
+	$DB_MODEL -> {tables} -> {$table_name} or return;
+	
 	return if $DB_MODEL -> {tables} -> {$table_name} -> {keys} -> {fake};
 	
-	$model_update -> assert (tables => {
+	$model_update -> assert (
+	
+		tables => {
+	
+			$table_name => {
+				keys => {fake => 'fake'},
+			},
 	
-		$table_name => {
-			keys => {fake => 'fake'},
 		},
+		
+		prefix => 'assert_fake_key#',
+
+	);
+
+}
+
+################################################################################
+
+sub is_recyclable {
+
+	my ($table_name) = @_;
 	
-	},core_voc_replacement_use => $conf -> {core_voc_replacement_use});
+	return 0 if $table_name eq $conf -> {systables} -> {log};
+	return 0 if $table_name eq $conf -> {systables} -> {sessions};
+	
+	if (ref $conf -> {core_recycle_ids} eq ARRAY) {
+		$conf -> {core_recycle_ids} = {map {$_ => 1} @{$conf -> {core_recycle_ids}}}
+	}
+
+	return 1 if $conf -> {core_recycle_ids} == 1 || $conf -> {core_recycle_ids} -> {$table_name};
+	return 0;
 
 }
 
@@ -814,11 +925,11 @@ sub delete_fakes {
 	
 	$table_name    ||= $_REQUEST {type};
 
-	return if is_recyclable ($table_name);
+	return if ($_REQUEST {__delete_fakes} -> {$table_name} ||= is_recyclable ($table_name));
 	
-	assert_fake_key ($table_name,core_voc_replacement_use => $conf -> {core_voc_replacement_use});
+	assert_fake_key ($table_name);
 	
-	my $ids = sql_select_ids (<<EOS);
+	my ($ids, $in_clause) = sql_select_ids (<<EOS);
 		SELECT
 			$table_name.id
 		FROM
@@ -829,7 +940,9 @@ sub delete_fakes {
 			AND $conf->{systables}->{sessions}.id_user IS NULL
 EOS
 			
-	sql_do ("DELETE FROM $table_name WHERE id IN ($ids)");
+	sql_do ("DELETE FROM $table_name WHERE id IN ($in_clause)");
+	
+	$_REQUEST {__delete_fakes} -> {$table_name} = 1;
 
 }
 
@@ -931,7 +1044,7 @@ sub sql_extract_params {
 			) {
 
 				my $value = $1;
-				$value =~ s{\\\'}{\'}gsm;
+				$value =~ s{\\\'}{\'}gsm; #'
 				push @params1, $value;
 				$token = '?';
 
@@ -985,9 +1098,8 @@ sub __log_request_profilinig {
 
 	my ($request_time) = @_;
 
-	return 
-		unless ($preconf -> {core_debug_profiling} > 2 && $model_update -> {core_ok});
-		
+	return unless ($preconf -> {core_debug_profiling} > 2 && $model_update -> {core_ok});
+
 	my $c = $r -> connection; 
 
 	$_REQUEST {_id_request_log} = sql_do_insert ($conf -> {systables} -> {__request_benchmarks}, {
@@ -996,13 +1108,15 @@ sub __log_request_profilinig {
 		ip_fw	=> $ENV {HTTP_X_FORWARDED_FOR},
 		fake	=> 0,
 		type	=> $_REQUEST {type},
-		mac	=> (!$preconf -> {core_no_log_mac}) ? get_mac () : '',
+		mac	=> get_mac (),
 		request_time	=> int ($request_time),
 		connection_id	=> $c -> id (),
 		connection_no	=> $c -> keepalives (),
 	});
 	
-	sql_do ("UPDATE $conf->{systables}->{__request_benchmarks} SET params = ? WHERE id = ?",
+	my $request_benchmarks_table = sql_table_name ($conf -> {systables} -> {__request_benchmarks});
+
+	sql_do ("UPDATE $request_benchmarks_table SET params = ? WHERE id = ?",
 		Data::Dumper -> Dump ([\%_REQUEST], ['_REQUEST']), $_REQUEST {_id_request_log}); 
 
 }
@@ -1018,7 +1132,9 @@ sub __log_request_finish_profilinig {
 
 	my $time = time;
 
-	sql_do ("UPDATE $conf->{systables}->{__request_benchmarks} SET application_time = ?, sql_time = ?, response_time = ?, bytes_sent = ?, is_gzipped = ? WHERE id = ?",
+	my $request_benchmarks_table = sql_table_name ($conf -> {systables} -> {__request_benchmarks});
+
+	sql_do ("UPDATE $request_benchmarks_table SET application_time = ?, sql_time = ?, response_time = ?, bytes_sent = ?, is_gzipped = ? WHERE id = ?",
 		int ($options -> {application_time}), 
 		int ($options -> {sql_time}), 
 		$options -> {out_html_time} ? int (1000 * (time - $options -> {out_html_time})) : 0, 
@@ -1026,5 +1142,385 @@ sub __log_request_finish_profilinig {
 		$options -> {is_gzipped},		 
 		$options -> {id_request_log},
 	);
+
 }
-1;
+
+################################################################################
+
+sub sql_select_ids {
+
+	my ($sql, @params) = @_;	
+
+	my $ids;
+
+	my $tied = tie $ids, 'Eludia::Tie::IdsList', {
+	
+		sql 			=> $sql,
+		
+		_REQUEST 		=> \%_REQUEST,
+		
+		package 		=> __PACKAGE__,
+		
+		params 			=> \@params,
+		
+		db 			=> $db,
+
+		sql_translator_ref	=> get_sql_translator_ref(),
+
+		
+	};
+	
+	return wantarray ? (
+		$ids,
+		wantarray && _sql_ok_subselects () ? $tied -> _sql : $ids,
+	) : $ids;
+
+}
+
+################################################################################
+
+sub sql_upload_files {
+
+	my ($options) = @_;
+	
+	my @nos = ();
+	
+	foreach my $k (keys %_REQUEST) {
+
+		$k =~ /^_$options->{name}_(\d+)$/ or next;
+		
+		$_REQUEST {$k} or next;
+		
+		push @nos, $1;
+
+	}
+
+	@nos > 0 or return;
+
+	my ($table, $field) = split /\./, $_REQUEST {"__$options->{name}_file_field"};
+	
+	$options -> {id} ||= $_REQUEST {id};
+	
+	sql_do ("UPDATE $table SET fake = -1 WHERE $field = ?", $options -> {id});
+	
+	my $name = $options -> {name};
+	
+	my $id = $options -> {id};
+	
+	$options -> {table}            = $table;
+	$options -> {file_name_column} = 'file_name';
+	$options -> {size_column}      = 'file_size';
+	$options -> {type_column}      = 'file_type';
+	$options -> {path_column}      = 'file_path';
+	$options -> {body_column}      = 'file_body' if $model_update -> get_columns ($table) -> {file_body};
+
+	foreach my $no (sort {$a <=> $b} @nos) {
+		
+		$options -> {name} = "${name}_${no}";
+
+		$options -> {id} = sql_do_insert ($table => {
+
+			$field => $id,
+			fake   => 0,
+			
+		});
+		
+		sql_upload_file ($options);
+	
+	}
+
+	sql_select_loop ("SELECT * FROM $table WHERE $field = ? AND fake = -1", sub {
+	
+		my $path = $i -> {$options -> {path_column}} or return;
+		
+		unlink $r -> document_root . $path;
+	
+	}, $id);
+
+	sql_do ("DELETE FROM $table WHERE $field = ? AND fake = -1", $id);
+
+}
+
+################################################################################
+################################################################################
+
+#package DBIx::ModelUpdate;
+
+use DBI::Const::GetInfoType;
+
+################################################################################
+
+sub new {
+
+	my ($package_name, $db, @options) = @_;
+	
+	my $driver_name = $db -> get_info ($GetInfoType {SQL_DBMS_NAME});
+	
+	$driver_name =~ s{\s}{}gsm;
+	
+	die $@ if $@;
+
+	my $self = bless ({
+		db => $db, 
+		driver_name => $driver_name,
+		quote => $db -> get_info ($GetInfoType {SQL_IDENTIFIER_QUOTE_CHAR}),
+		@options
+	}, $package_name);
+	
+	if ($driver_name eq 'Oracle') {
+  		$self -> {characterset} = sql_select_scalar ('SELECT VALUE FROM V$NLS_PARAMETERS WHERE PARAMETER = ?', 'NLS_CHARACTERSET');
+  		$self -> {schema} ||= uc $db -> {Username};
+	}
+	
+	$self -> {schema} ||= '';
+
+	return $self;
+
+}
+
+################################################################################
+
+sub sql_assert_default_columns {
+
+	my ($needed_tables, $params) = @_;
+
+	my $default_columns = $params -> {default_columns} or return $needed_tables;
+
+	foreach my $name (keys %$needed_tables) {
+	
+		my $definition = $needed_tables -> {$name};
+
+		next if $definition -> {sql};
+
+		next if $definition -> {columns} -> {id};
+
+		foreach my $dc_name (keys %$default_columns) {
+
+			$definition -> {columns} -> {$dc_name} ||= Storable::dclone $default_columns -> {$dc_name};
+
+		}
+
+	}
+	
+	return $needed_tables;
+
+}
+
+################################################################################
+
+sub assert {
+
+	my ($self, %params) = @_;
+	
+	my $core_debug_sql_do = $preconf -> {core_debug_sql_do};
+
+	$preconf -> {core_debug_sql_do} = 1;
+
+	my ($tables, my $new_checksums) = checksum_filter (db_model => $params {prefix}, 
+	
+		sql_assert_default_columns (Storable::dclone ($params {tables}), \%params)
+		
+	);
+		
+	my $objects = [\my @tables, \my @views];
+
+	while (my ($name, $object) = each %$tables) {
+	
+		next if $object -> {off};
+	
+		$object -> {name} = $name;
+
+		push @{$objects -> [$object -> {sql} ? 1 : 0]}, $object;
+
+	}
+
+	wish (tables => Storable::dclone \@tables, {});
+
+	foreach my $table (@tables) {
+
+		wish (table_columns => [map {{name => $_, %{$table -> {columns} -> {$_}}}}    (keys %{$table -> {columns}})], {table => $table -> {name}}) if exists $table -> {columns};
+
+		wish (table_keys    => [map {{name => $_, parts => $table -> {keys} -> {$_}}} (keys %{$table -> {keys}})],    {table => $table -> {name}, table_def => $table}) if exists $table -> {keys};
+		
+		if (exists $table -> {data} && ref $table -> {data} eq ARRAY && @{$table -> {data}} > 0) {
+				
+			wish (table_data => $table -> {data}, {
+			
+				table => $table -> {name},
+				
+				key   => exists $table -> {data} -> [0] -> {id} ? 'id' : 'name',
+				
+			});
+		
+		}
+
+	}
+	
+	wish (views => \@views, {});
+
+	$preconf -> {core_debug_sql_do} = $core_debug_sql_do;
+
+	checksum_write ('db_model', $new_checksums);
+
+}
+
+#############################################################################
+
+sub sql_store_ids {
+
+	my $options;
+	
+	if (@_ == 2 && !ref $_[0] && !ref $_[1]) {
+	
+		$options = {
+			table => $_[0],
+			key   => $_[1],
+		}
+	
+	}
+	elsif (ref $_[0] eq HASH) {
+	
+		$options = $_[0];
+		
+	}
+	else {
+	
+		die "Wrong parameters for sql_store_ids: " . Dumper (\@_);
+	
+	}
+	
+	$options -> {root} ||= {'id_' . en_unplural ($_REQUEST {type}) => $_REQUEST {id}};
+	
+	wish (table_data =>	[map {{	fake => 0, $options -> {key} => $_ }} get_ids ($options -> {key})], $options);
+
+}
+
+#############################################################################
+
+sub sql_clone {
+
+	my ($table, $data, %fields) = @_;
+
+	my $clone = {%$data, %fields};
+
+	delete $clone -> {id};
+
+	$clone -> {id} = sql_do_insert ($table => $clone);
+
+	return $clone;
+
+}
+
+#############################################################################
+
+sub require_wish ($) {
+
+	return if $INC_FRESH {"Wish::$_[0]"};
+	
+	foreach my $key (map {"Eludia/SQL$_/Wish/$_[0].pm"} ('', '/Dialect/' . $SQL_VERSION -> {driver})) {
+	
+		eval {require $key};
+		
+		delete $INC {$key};
+
+	}
+
+	$INC_FRESH {"Wish::$_[0]"} = 1;
+
+}
+
+#############################################################################
+
+sub wish {
+
+	my ($type, $items, $options) = @_;
+
+	require_wish $type;
+
+	&{"wish_to_adjust_options_for_$type"} ($options);
+		
+	foreach my $i (@$items) { &{"wish_to_clarify_demands_for_$type"} ($i, $options) }
+
+	my $existing = &{"wish_to_explore_existing_$type"} ($options);
+
+	my $todo = {};
+	
+	my @key = @{$options -> {key}};
+
+	foreach my $new (@$items) {
+
+		my $old = delete $existing -> {join '_', @$new {@key}} or (push @{$todo -> {create}}, $new) and next;
+
+		&{"wish_to_update_demands_for_$type"} ($old, $new, $options);
+
+		next if Dumper ($new) eq Dumper ($old);
+
+		&{"wish_to_schedule_modifications_for_$type"} ($old, $new, $todo, $options);		
+		
+	}
+
+	&{"wish_to_schedule_cleanup_for_$type"} ($existing, $todo, $options);
+
+	foreach my $action (keys %$todo) { &{"wish_to_actually_${action}_${type}"} ($todo -> {$action}, $options) }
+
+}
+
+#############################################################################
+
+sub get_tables {
+
+	my ($self, $table) = @_;
+
+	require_wish 'tables';
+		
+	return sort keys %{wish_to_explore_existing_tables ()};
+
+}
+
+#############################################################################
+
+sub get_columns {
+
+	my ($self, $table) = @_;
+
+	require_wish 'table_columns';
+	
+	wish_to_adjust_options_for_table_columns (my $options = {table => $table});
+	
+	return wish_to_explore_existing_table_columns ($options);
+
+}
+
+#############################################################################
+
+sub get_keys {
+
+	my ($self, $table) = @_;
+	
+	require_wish 'table_keys';
+	
+	wish_to_adjust_options_for_table_keys (my $options = {table => $table});
+	
+	my %keys = ();
+		
+	foreach my $i (values %{wish_to_explore_existing_table_keys ($options)}) {		
+		
+		if ($i -> {global_name} =~ /.*?$options->{table}_/i) {
+		
+			$i -> {global_name} = $';
+		
+		}
+			
+		$keys {lc $i -> {global_name}} = join ', ', @{$i -> {parts}}
+
+	}
+		
+	return \%keys;
+
+}
+
+#############################################################################
+
+sub sql_table_name {$_[0]}
+
+1;
\ No newline at end of file
